"""
Dynamically allocated van Emde Boas Trees (Queues).

"""

from collections.abc import MutableSet
from math import ceil, floor, log2


class _EmptyVEBTree:
    min = max = None
    universe_size = 0

    def __contains__(self, i):
        return False

    def __iter__(self):
        return iter([])

    def __len__(self):
        return 0

    def predecessor(self, i):
        return None

    def successor(self, i):
        return None

    def discard(self, i):
        return None


_EMPTY = _EmptyVEBTree()


class vEBTree(MutableSet):
    _root = _EMPTY
    universe_size = _root.universe_size
    discard = _root.discard
    predecessor = _root.predecessor
    successor = _root.successor

    def __init__(self, contents=()):
        if contents:
            self.update(contents)

    def __bool__(self):
        return self.min is not None

    __nonzero__ = __bool__

    def __contains__(self, i):
        return i in self._root

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return (
            self.universe_size == other.universe_size
            and self.min == other.min
            and self.max == other.max
            and len(self) == len(other)
            and all(i == j for i, j in zip(self, other))
        )

    def __iter__(self):
        return iter(self._root)

    def __len__(self):
        return len(self._root)

    def __repr__(self):
        return "vEBTree(%r)" % list(self)

    def _update_root(self, new_root):
        self._root = new_root
        self.universe_size = new_root.universe_size
        self.discard = new_root.discard
        self.predecessor = new_root.predecessor
        self.successor = new_root.successor

    @classmethod
    def of_size(cls, n):
        tree = cls()
        tree.grow(n)
        return tree

    @property
    def min(self):
        return self._root.min

    @property
    def max(self):
        return self._root.max

    def grow(self, to_size):
        if to_size <= self.universe_size:
            return
        if to_size <= 2:
            return self._update_root(_vEBLeaf())

        # u = 2 ^ m -> m = log2(u), u = to_size, m = square_root
        square_root = ceil(log2(to_size))
        old_root = self._root
        old_size = self.universe_size
        new_size = 1 << square_root
        # 1 << m = 2 ^ m
        self._update_root(_vEBTree(new_size))

        if old_root is not _EMPTY:
            # when tree grows from even powers of 2 into odd ones, for example
            # 4 into 8 (2^2 into 2^3), sizes of clusters stay the same, which
            # allows to copy them directly from old root.
            if new_size - old_size == 1 and old_size % 2 == 0:
                for i in range(len(old_root.clusters)):
                    self._root.clusters[i] = old_root.clusters[i]
                # updating clusters means updating summary
                # NOTE: this worked first time even though I didn't really
                # know how exactly summary works. be cautious here.
                for i in range(len(old_root.summary)):
                    self._root.summary[i] = old_root.summary[i]
                # min and max properties need to be updated
                self._root.min = old_root.min
                self._root.max = old_root.max
            else:
                self.update(old_root)

    def add(self, i):
        if i >= self.universe_size:
            self.grow(i + 1)
        self._root.add(i)

    def update(self, iterable):
        maxX = -1
        for x in iterable:
            maxX = max(maxX, x)
        # this is done to prevent multiple grows per one update call
        if maxX >= self.universe_size:
            self.grow(maxX + 1)
        # adding maxX twice doesn't hurt anybody, so it's not excluded
        for i in iterable:
            self.add(i)


class _vEBLeaf:
    universe_size = 2

    def __init__(self):
        self.values = [False, False]

    def __contains__(self, x):
        if x > 2:
            return False
        return self.values[x]

    def __iter__(self):
        return (i for i, v in enumerate(self.values) if v)

    def __len__(self):
        return self.values.count(True)

    def __reversed__(self):
        return (i for i, v in zip((1, 0), reversed(self.values)) if v)

    @property
    def min(self):
        return next(iter(self), None)

    @property
    def max(self):
        return next(reversed(self), None)

    def add(self, x):
        self.values[x] = True

    def discard(self, x):
        self.values[x] = False

    def predecessor(self, x):
        if x < 1:
            return
        elif x == 1:
            return 0 if self.values[0] else None
        else:
            return 1 if self.values[1] else 0 if self.values[0] else None

    def successor(self, x):
        if x > 0:
            return
        elif x == 0:
            return 1 if self.values[1] else None
        else:
            return 0 if self.values[0] else 1 if self.values[1] else None


class _vEBTree:
    min = max = None

    def __init__(self, n):
        root = ceil(log2(n)) / 2
        upper, self._lower = 1 << ceil(root), 1 << floor(root)
        self.summary = vEBTree.of_size(upper)
        self.clusters = [None] * upper
        self.universe_size = n

    def __contains__(self, x):
        if self.min is None:
            return False
        elif x == self.min:
            return True

        high, low = divmod(x, self._lower)
        if self.clusters[high] is None:
            return False
        return low in self.clusters[high]

    def __iter__(self):
        i, _ = minimum, maximum = self.min, self.max

        if minimum is None:
            return
        yield minimum

        while i != maximum:
            i = self.successor(i)
            yield i

    def __reversed__(self):
        _, i = minimum, maximum = self.min, self.max

        if minimum is None:
            return
        yield maximum

        while i != minimum:
            i = self.predecessor(i)
            yield i

    def __len__(self):
        return sum(1 for _ in self)

    def add(self, x):
        if self.min is None:
            self.min = self.max = x
            return

        if x == self.min:
            return
        if x < self.min:
            x, self.min = self.min, x
        if x > self.max:
            self.max = x

        high, low = divmod(x, self._lower)
        cluster = self.clusters[high]

        if cluster is None:
            cluster = self.clusters[high] = vEBTree.of_size(
                self.summary.universe_size,
            )
            self.summary.add(high)
        cluster.add(low)

    def discard(self, x):
        if self.min is None or x < self.min:
            return

        if x == self.min:
            new_min_in = self.summary.min
            if new_min_in is None:
                self.min = self.max = None
            else:
                cluster = self.clusters[new_min_in]
                new_min = cluster.min
                x = self.min = new_min_in * self._lower + new_min
                # XXX: I don't know why just changing to
                # cluster.discard(new_min_in) and returning doesn't work here

        high, low = divmod(x, self._lower)
        cluster = self.clusters[high]

        if cluster is None:
            return
        cluster.discard(low)

        if cluster.min is None:
            self.clusters[high] = None
            self.summary.discard(high)

        if x == self.max:
            global_max = self.summary.max
            if global_max is None:
                self.max = self.min
            else:
                self.max = (
                    global_max * self._lower + self.clusters[global_max].max
                )

    def predecessor(self, x):
        if self.min is None or x <= self.min:
            return None
        elif x > self.max:
            return self.max

        high, low = divmod(x, self._lower)
        cluster = self.clusters[high]

        if cluster is None or low <= cluster.min:
            high = self.summary.predecessor(high)
            if high is None:
                return self.min
            return high * self._lower + self.clusters[high].max
        else:
            low = cluster.predecessor(low)
            return high * self._lower + low

    def successor(self, x):
        if self.min is None or x >= self.max:
            return None
        elif x < self.min:
            return self.min

        high, low = divmod(x, self._lower)
        cluster = self.clusters[high]

        if cluster is None or low >= cluster.max:
            high = self.summary.successor(high)
            return high * self._lower + self.clusters[high].min
        else:
            low = cluster.successor(low)
            return high * self._lower + low



from unittest import TestCase, expectedFailure
import bisect
import random

from veb import vEBTree


class VEBTestMixin:
    def test_repr(self):
        self.assertEqual(repr(self.t), "vEBTree(%r)" % list(self.t))

    def test_len(self):
        self.assertEqual(len(self.t), 0)

        self.t.add(0)
        self.assertEqual(len(self.t), 1)

        self.t.add(1)
        self.assertEqual(len(self.t), 2)

    def test_it_is_empty_when_created(self):
        for i in range(self.t.universe_size):
            self.assertNotIn(i, self.t)

        self.assertEqual(list(self.t), [])

    def test_iter_returns_contained_elements(self):
        evens = range(self.t.universe_size, 0, 2)
        for i in evens:
            self.t.add(i)
        self.assertEqual(list(self.t), list(evens))

    def test_bool_returns_whether_empty(self):
        self.assertFalse(self.t)
        self.t.add(0)
        self.assertTrue(self.t)

    def test_out_of_bounds_containment_is_false(self):
        self.assertNotIn(self.t.universe_size + 1, self.t)

    def test_add_twice_does_nothing(self):
        self.t.add(0)
        self.t.add(0)
        self.assertIn(0, self.t)

    def test_discard_not_present_does_nothing(self):
        self.t.discard(0)
        self.t.discard(0)
        self.assertNotIn(0, self.t)

    def test_negatives(self):
        self.t.update([0, 1])
        self.assertEqual(self.t.successor(-1), 0)
        self.assertEqual(self.t.predecessor(self.t.universe_size + 1), 1)


class TestVEBTree(TestCase, VEBTestMixin):
    def setUp(self):
        self.t = vEBTree.of_size(4)

    def test_add_to_empty(self):
        self.t.add(3)
        self.assertNotIn(0, self.t)
        self.assertNotIn(1, self.t)
        self.assertNotIn(2, self.t)
        self.assertIn(3, self.t)

        self.assertEqual(self.t.min, 3)

    def test_add_min(self):
        self.t.add(0)
        self.assertIn(0, self.t)

        self.t.add(0)
        self.assertIn(0, self.t)

    def test_add_less_than_min(self):
        self.t.add(1)
        self.t.add(0)
        self.assertIn(0, self.t)
        self.assertIn(1, self.t)

        self.assertEqual(self.t.min, 0)

    def test_add_greater_than_max(self):
        self.t.add(0)
        self.t.add(1)
        self.assertIn(0, self.t)
        self.assertIn(1, self.t)

        self.assertEqual(self.t.max, 1)

    def test_discard_less_than_min(self):
        self.t.add(1)
        self.t.discard(0)
        self.assertNotIn(0, self.t)

        self.assertEqual(self.t.min, 1)

    def test_discard_min_last_element(self):
        self.t.add(1)
        self.t.discard(1)
        self.assertNotIn(1, self.t)
        self.assertIsNone(self.t.min)
        self.assertIsNone(self.t.max)

    def test_discard_min_update_to_new_min(self):
        self.t.update([0, 1])
        self.t.discard(0)
        self.assertNotIn(0, self.t)
        self.assertEqual(self.t.min, 1)
        self.assertEqual(self.t.max, 1)

    def test_predecessor_empty(self):
        self.assertIsNone(self.t.predecessor(2))

    def test_predecessor_greater_than_max(self):
        self.t.add(2)
        self.assertEqual(self.t.predecessor(3), 2)

    def test_predecessor_min_or_smaller(self):
        self.t.add(3)
        self.assertEqual(self.t.min, 3)
        self.assertIsNone(self.t.predecessor(3))

    def test_predecessor_in_current_cluster(self):
        self.t.update([0, 3])
        self.assertEqual(self.t.predecessor(1), 0)

    def test_predecessor_in_next_cluster_current_cluster_only_has_min(self):
        self.t.update([0, 3])
        self.assertEqual(self.t.predecessor(2), 0)

    def test_predecessor_in_next_cluster_current_cluster_has_elements(self):
        self.t.update([0, 2, 3])
        self.assertEqual(self.t.predecessor(2), 0)

    def test_successor_empty(self):
        self.assertIsNone(self.t.successor(2))

    def test_successor_less_than_min(self):
        self.t.add(3)
        self.assertEqual(self.t.successor(2), 3)

    def test_successor_max_or_bigger(self):
        self.t.add(3)
        self.assertEqual(self.t.max, 3)
        self.assertIsNone(self.t.successor(3))

    def test_successor_in_current_cluster(self):
        self.t.update([1, 3])
        self.assertEqual(self.t.successor(2), 3)

    def test_successor_in_next_cluster_current_cluster_only_has_min(self):
        self.t.update([1, 3])
        self.assertEqual(self.t.successor(1), 3)

    def test_successor_in_next_cluster_current_cluster_has_elements(self):
        self.t.update([0, 1, 3])
        self.assertEqual(self.t.successor(1), 3)

    def test_grow_on_addition(self):
        t = vEBTree.of_size(2)
        self.assertEqual(t.universe_size, 2)

        t.add(2)
        self.assertEqual(t.universe_size, 4)
        self.assertIn(2, t)
        self.assertNotIn(3, t)

        t.add(7)
        self.assertEqual(t.universe_size, 8)
        self.assertIn(2, t)
        self.assertNotIn(3, t)
        self.assertIn(7, t)
        self.assertNotIn(6, t)

        t.add(195)
        self.assertEqual(t.universe_size, 256)

    def test_grow_keeps_elements(self):
        t = vEBTree.of_size(8)
        added = [2, 7, 5]
        for x in added:
            t.add(x)
        t.grow(32)
        saved = list(t)
        # there should be equal amount of added and saved distinct elements
        self.assertEqual(len(added), len(saved))
        # every added element should have been saved
        added.sort()
        saved.sort()
        for i in range(len(added)):
            self.assertEqual(added[i], saved[i])

    def test_grow_keeps_max(self):
        t = vEBTree.of_size(4)

        t.add(2)
        self.assertEqual(t.max, 2)

        t.grow(16)
        self.assertEqual(t.max, 2)

    def test_grow_keeps_min(self):
        t = vEBTree.of_size(4)

        t.add(2)
        self.assertEqual(t.min, 2)

        t.grow(16)
        self.assertEqual(t.min, 2)

    def test_of_size(self):
        t = vEBTree.of_size(16)
        self.assertEqual(t.universe_size, 16)

    def test_update(self):
        # test 1
        self.t.update((1, 2))
        self.assertIn(1, self.t)
        self.assertIn(2, self.t)
        # test 2
        e = []
        for x in range(32):
            e.append(x)
        t = vEBTree([])
        t.update(e)
        self.assertEqual(len(t), len(e))
        for x in e:
            self.assertIn(x, t)

    def test_init_adds_all_contents(self):
        t = vEBTree([1, 3, 15])
        self.assertEqual(t.universe_size, 16)
        self.assertIn(1, t)
        self.assertIn(3, t)
        self.assertIn(15, t)

    def test_equal_if_same_contents(self):
        t = vEBTree.of_size(self.t.universe_size)
        self.assertEqual(t, self.t)

        self.t.add(2)
        self.assertNotEqual(t, self.t)

        t.add(2)
        self.assertEqual(t, self.t)

        t.update([3, 5, 7])
        self.assertNotEqual(t, self.t)

        self.t.update([3, 5, 7])
        self.assertEqual(t, self.t)


class TestSize2VEBTree(TestCase, VEBTestMixin):
    def setUp(self):
        self.t = vEBTree.of_size(2)

    def test_add(self):
        self.t.add(0)
        self.assertIn(0, self.t)
        self.assertNotIn(1, self.t)

        self.t.add(1)
        self.assertIn(0, self.t)
        self.assertIn(1, self.t)

    def test_discard(self):
        self.assertNotIn(0, self.t)
        self.t.discard(0)
        self.assertNotIn(0, self.t)

        self.t.add(0)
        self.t.discard(0)
        self.assertNotIn(0, self.t)

    def test_min_max(self):
        self.assertIsNone(self.t.min)
        self.assertIsNone(self.t.max)

        self.t.add(0)
        self.assertEqual(self.t.min, 0)
        self.assertEqual(self.t.max, 0)

        self.t.add(1)
        self.assertEqual(self.t.min, 0)
        self.assertEqual(self.t.max, 1)

    def test_predecessor(self):
        self.assertIsNone(self.t.predecessor(0))
        self.assertIsNone(self.t.predecessor(1))

        self.t.add(0)
        self.assertIsNone(self.t.predecessor(0))
        self.assertEqual(0, self.t.predecessor(1))

    def test_successor(self):
        self.assertIsNone(self.t.successor(0))
        self.assertIsNone(self.t.successor(1))

        self.t.add(1)
        self.assertEqual(1, self.t.successor(0))
        self.assertIsNone(self.t.successor(1))


class TestEmptyVEBTree(TestCase, VEBTestMixin):
    def setUp(self):
        self.t = vEBTree()

    def test_size_0_makes_empty_tree(self):
        t = vEBTree.of_size(0)
        self.assertEqual(t.universe_size, 0)

    def test_universe_size(self):
        self.assertEqual(self.t.universe_size, 0)

    def test_repr(self):
        self.assertEqual(repr(self.t), "vEBTree([])")

    def test_bool(self):
        self.assertFalse(self.t)

    def test_contains(self):
        self.assertNotIn(0, self.t)

    def test_iter(self):
        self.assertEqual(list(self.t), [])

    def test_len(self):
        self.assertEqual(len(self.t), 0)

    def test_min(self):
        self.assertIsNone(self.t.min)

    def test_max(self):
        self.assertIsNone(self.t.max)

    def test_predecessor(self):
        self.assertIsNone(self.t.predecessor(0))

    def test_successor(self):
        self.assertIsNone(self.t.successor(0))

    def test_grow(self):
        self.t.grow(2)
        self.assertEqual(self.t.universe_size, 2)


class VEBQueueTest:
    @expectedFailure
    def testCreateNotEvenPowerOfTwo(self):
        self.assertRaises(IndexError, vEBTree.of_size, 0)
        self.assertRaises(IndexError, vEBTree.of_size, 3)


class LameQueue:
    def __init__(self):
        self.q = []

    def __contains__(self, x):
        locate = bisect.bisect_left(self.q, x)
        if locate == len(self.q) or self.q[locate] != x:
            return False
        return True

    def add(self, x):
        # Do a sorted insert
        insertion_point = bisect.bisect_left(self.q, x)
        if insertion_point == len(self.q) or self.q[insertion_point] != x:
            self.q.insert(insertion_point, x)

    def discard(self, x):
        insertion_point = bisect.bisect_left(self.q, x)
        if insertion_point < len(self.q) and self.q[insertion_point] == x:
            del self.q[insertion_point]

    def max(self):
        if len(self.q) > 0:
            return self.q[-1]
        return None

    def min(self):
        if len(self.q) > 0:
            return self.q[0]
        return None

    def predecessor(self, x):
        locate = bisect.bisect_left(self.q, x) - 1
        if locate >= 0:
            return self.q[locate]
        return None


class RandomTest(TestCase):
    def testRandom(self):
        n = 1 << 16
        totalOperations = 1 << 16
        operations = totalOperations

        q = vEBTree.of_size(n)
        lame = LameQueue()
        while operations > 0:
            # Do some searches
            numSearches = random.randint(0, 10)
            operations -= numSearches
            for _ in range(numSearches):
                search = random.randint(0, n - 1)
                r1 = search in lame
                r2 = search in q
                self.assertEqual(r1, r2)

                r1 = lame.predecessor(search)
                r2 = q.predecessor(search)
                self.assertEqual(r1, r2)

            # Change the queue state
            operations -= 1
            value = random.randint(0, n - 1)
            stateChange = random.randint(0, 1)
            if stateChange == 0:
                q.add(value)
                lame.add(value)
                #  print "add", value
            else:
                q.discard(value)
                lame.discard(value)
                #  print "remove", value

            # Check the max/min variables
            assert lame.max() == q.max
            assert lame.min() == q.min



from pathlib import Path

import nox

ROOT = Path(__file__).parent
PYPROJECT = ROOT / "pyproject.toml"
DOCS = ROOT / "docs"
PACKAGE = ROOT / "veb"


nox.options.sessions = []


def session(default=True, **kwargs):
    def _session(fn):
        if default:
            nox.options.sessions.append(kwargs.get("name", fn.__name__))
        return nox.session(**kwargs)(fn)

    return _session


@session(python=["3.8", "3.9", "3.10", "3.11", "pypy3"])
def tests(session):
    dependencies = ["virtue", ROOT]
    if session.posargs == ["coverage"]:
        dependencies.append("coverage[toml]")
        session.install(*dependencies)
        session.run("coverage", "run", "-m", "virtue", PACKAGE)
        session.run("coverage", "report")
    else:
        session.install(*dependencies)
        session.run("virtue", *session.posargs, PACKAGE)


@session(tags=["build"])
def build(session):
    session.install("build")
    tmpdir = session.create_tmp()
    session.run("python", "-m", "build", ROOT, "--outdir", tmpdir)


@session(tags=["style"])
def readme(session):
    session.install("build", "twine")
    tmpdir = session.create_tmp()
    session.run("python", "-m", "build", ROOT, "--outdir", tmpdir)
    session.run("python", "-m", "twine", "check", tmpdir + "/*")


@session(tags=["style"])
def style(session):
    session.install(
        "flake8",
        "flake8-broken-line",
        "flake8-bugbear",
        "flake8-commas",
        "flake8-quotes",
        "flake8-tidy-imports",
    )
    session.run("python", "-m", "flake8", PACKAGE, __file__)



from random import randint
from timeit import timeit

from veb import vEBTree


def random_tree(size):
    tree = vEBTree.of_size(size)
    for _ in range(randint(0, tree.universe_size - 1)):
        tree.add(randint(0, tree.universe_size - 1))
    return tree


for size in range(1, 16):
    bench = "random_tree(2 ** {})".format(size)
    t = timeit(
        bench, "from __main__ import random_tree, size, vEBTree", number=1000
    )
    print(bench, ":", t)
